#!/usr/bin/python

# libraries
import serial
import time
import datetime
import sys
import pifacedigitalio as pfio
import binascii
import os
 
# settings
PRINT = False
KLINEOUT = 0	# PiFace i/o number for the k-line relay
TIMESTART = 0.5 # quiet time before starting communications
TIMEKLOW = 0.070 # Kline is pulled low for this time
TIMEKHIGH = 0.20 # quiet time after low pulse before wakeup
TIMEINIT = 0.2 # time between wakeup and initialise
TIMEBTWMESS = 0.1 # time between messages
TIMEWAITRX = 0.2 # time between sending a message and listening to receive a response
TIMEUPDATE = 0.1 # time between update cycles
ERRORMAX = 4 # number of consecutive read errors that triggers a re-start 
COUNTMAX = 10 # maximum number of blank port reads per message
CPUPAUSE = 0.02 # slows down the read loop for multitasking
GEARTABLE = [117, 79, 60, 50, 43, 35] # rpm per kph plus 5%
FUELMIN = 3.5 # the minimum fuel below which the fuel level is counted down by the ecu

# Messages and expected response lengths
WAKEUP = 'FE 04 FF FF' # no response expected
WAKEUPLEN = 0
INITIALISE = '72 05 00 F0 99' # initialise communications
INITLEN = 4
REQTABLE11 = '72 07 72 11 00 14 F0' # request all of table 11 data
TABLE11LEN = 26 # length in decimal
REQTABLE00 = '72 05 71 00 18'
REQTABLE20 = '72 05 71 20 F8'
REQTABLED0 = '72 05 71 D0 48'

# Files
	# controlled by the web interface
CONTROLFILE = '/var/www/obdcontrol/status.txt'
KFUELFILE = '/var/www/obdcontrol/kfuel.txt'
ALPHAFILE = '/var/www/obdcontrol/alpha.txt'
BETAFILE = '/var/www/obdcontrol/beta.txt'

	# controlled by this program
LOGFILE = '/home/pi/obd/log.txt' 
FUELFILE = '/home/pi/obd/fuel.txt'
FUELDELTAFILE = '/home/pi/obd/fueldelta.txt'
SPEEDFILE = '/home/pi/obd/speed.txt'
KLAFILE = '/home/pi/obd/kla.txt'
KLIFILE = '/home/pi/obd/kli.txt'
ECTFILE = '/home/pi/obd/ect.txt'
IATFILE = '/home/pi/obd/iat.txt'
GEARFILE = '/home/pi/obd/gear.txt'
VOLTFILE = '/home/pi/obd/voltage.txt'
INJECTORFILE = '/home/pi/obd/injector.txt'
SCANTIMEFILE = '/home/pi/obd/scantime.txt'
RPMFILE = '/home/pi/obd/rpm.txt'
COMMFILE = '/home/pi/obd/commerror.txt'
 
# initial values 
RPM = 0
VOLTAGE = 0.0
KLI = 0.0
KLA = 0.0
INJ = 0.0
SPEED = 0
GEAR = 0
TPS = 0
ECT = 0
IAT = 0
MAP = 0.0
FUEL = 0.0
FUELDELTA = 0.0

# environment
pfio.init()


# Functions

	# Wries error messages to the log
def errorlog(MESSAGE):
	if PRINT:
		print(MESSAGE)
        try:
                FILE = open(LOGFILE,'a')
                FILE.write(str(datetime.datetime.now()) + " " + MESSAGE + "\n")
                FILE.close()
        except:
                if PRINT:
                        print "Failed to write to LOG file"
        return

def flushrx():
        while ser0.inWaiting() >0:
               CLEAR = ser0.read(ser0.inWaiting()) # clear out any characters in the buffer
               CLEAR = binascii.hexlify(CLEAR).upper()
               CLEAR = ' '.join([CLEAR[i:i+2] for i in range(0, len(CLEAR), 2)])
               errorlog("< Cleared port: >" + CLEAR) # log any chars in buffer
               time.sleep(CPUPAUSE)
	ser0.flushOutput()

	return


	# writes a message to the serial port, and returns the response
def serialw(MESSAGE,LENGTH):
	flushrx()
	ERROR = False	
	SEND = binascii.unhexlify(MESSAGE.replace(" ",""))
	ser0.write(SEND)
	RESPONSE = ""
	COUNT = 0
	time.sleep(TIMEWAITRX)
	while len(RESPONSE) < len(SEND) + LENGTH:
		if ser0.inWaiting()>0:
			RESPONSE += ser0.read(1)
		else:
			COUNT += 1
			time.sleep(CPUPAUSE) # takes the CPU load down a bit
			if COUNT == COUNTMAX: # stops continuous cycling if expected chrs not recd
				errorlog("max count reached")
				break

	if PRINT:
		print "Count:", COUNT


	# parse the response
	RESPONSE = binascii.hexlify(RESPONSE).upper()
	RESPONSE = ' '.join([RESPONSE[i:i+2] for i in range(0, len(RESPONSE), 2)])
	
	if RESPONSE == "":
        	errorlog("Sent " + MESSAGE+ " < No echo or response >")
		ERROR = True
	else:
		if RESPONSE[:len(MESSAGE)] != MESSAGE: # check first part of response
			errorlog("< Message not echoed OK > : " + RESPONSE)
			ERROR = True
		else:
			# strip original message
			RESPONSE = RESPONSE.replace(MESSAGE,"")
			if RESPONSE == "" :
				errorlog("< Message echoed OK but no ECU response >")
				ERROR = True
			else: 
				RESPONSE = RESPONSE.lstrip()
				if len(RESPONSE) != LENGTH * 3 - 1:
					errorlog("< Incorrect response length >" + RESPONSE)
					ERROR = True
				else:
                        	       	# calculate checksum and compare checkdigit
					A = RESPONSE.split()	# convert to list
					B = int(A[-1],16)	# decimalise checkdigit
					A.pop()			# remove checkdigit
					C = 0
					for i in A: C = C + int(i,16)
					if B != 256 - C % 256:
						# checkdigit is bad
						errorlog(RESPONSE + " < Bad Checkdigit >")
						ERROR = True 

	if ERROR == True:
		RESPONSE = "" # return only good responses

	flushrx()
	return RESPONSE			

#------------------ Read/write inital values

try:
        FILE = open(KFUELFILE,'r')
        KFUEL = float(FILE.read())
        FILE.close()
except:
        errorlog("Failed to read KFUEL file")
        KFUEL = 900.0

try:
        FILE = open(FUELDELTAFILE,'r')
        FUELDELTA = float(FILE.read())
        FILE.close()
except:
        errorlog("Failed to read FUELDELTA file")
        FUELDELTA = 0.0

try:
        FILE = open(ALPHAFILE,'r')
        ALPHA = float(FILE.read())
        FILE.close()
except:
        errorlog("Failed to read ALPHA file")
        ALPHA = 0.25

try:
        FILE = open(BETAFILE,'r')
        BETA = float(FILE.read())
        FILE.close()
except:
        errorlog("Failed to read BETA file")
        BETA = 0.002

try:
        FILE = open(KLAFILE,'r')
        KLA = float(FILE.read())
        FILE.close()
except:
        errorlog("Failed to read KLA file")
        KLA = 30.0

try:
        FILE = open(RPMFILE,'w')
        FILE.write("0.0")
        FILE.close()
except:
        errorlog("Failed to write to RPM file")

try:
        FILE = open(SPEEDFILE,'w')
        FILE.write("0")
        FILE.close()
except:
        errorlog("Failed to write to SPEED file")

try:
        FILE = open(IATFILE,'w')
        FILE.write("0")
        FILE.close()
except:
        errorlog("Failed to write to IAT file")

try:
        FILE = open(ECTFILE,'w')
        FILE.write("0")
        FILE.close()
except:
        errorlog("Failed to write to ECT file")

try:
        FILE = open(VOLTFILE,'w')
        FILE.write("0.0")
        FILE.close()
except:
        if PRINT:
                print "Failed to write to VOLT file"

try:
        FILE = open(INJECTORFILE,'w')
        FILE.write("0")
        FILE.close()
except:
        errorlog("Failed to write to INJECTOR file")


# --------------- Main Program loop
FIRSTTIME = True
ERRORCOUNT = 0

while ERRORCOUNT < ERRORMAX:
	
	
	while True:
		STATUS = "stop"
		try:
			FILE = open(CONTROLFILE,'r')
			STATUS = FILE.read()
			FILE.close()
		except:
			errorlog("Failed to open STATUS file")

		if STATUS == "run":
			break
		try:
			FILE = open(COMMFILE, 'w')
			FILE.write("0")
			FILE.close()
		except:
			errorlog("Failed to write to COMMS file")

		sys.exit()

#--------------- Do initialisation on startup 
	if FIRSTTIME:
		FIRSTTIME = False
		errorlog("< STARTING >")
	        try:
        	        FILE = open(COMMFILE, 'r')
                	COMMS = int(FILE.read())
                	FILE.close()
                	FILE = open(COMMFILE, 'w')
                	FILE.write(str(COMMS + 1))
                	FILE.close()
        	except:
                	errorlog("Failed to write to COMMS file")

		ser0 = serial.Serial(
    		port='/dev/serial/by-id/usb-FTDI_FT232R_USB_UART_A50285BI-if00-port0',
    		baudrate=38400,
    		parity=serial.PARITY_NONE,
    		stopbits=serial.STOPBITS_ONE,
    		bytesize=serial.EIGHTBITS,
		dsrdtr=None,
		rtscts=None
		)

		if PRINT:
			print "Kline high "
		time.sleep(TIMESTART)
		if PRINT:
			print "Kline pulse"
		pfio.digital_write(KLINEOUT,1)
		time.sleep(TIMEKLOW)
		pfio.digital_write(KLINEOUT,0)
		time.sleep(TIMEKHIGH)


		DUMMY = serialw(WAKEUP, WAKEUPLEN) # send wakeup
		time.sleep(TIMEINIT)
		DUMMY = serialw(INITIALISE, INITLEN) # send initialise
		if DUMMY == "":
			ERRORCOUNT = ERRORMAX

# ---------------------- Interrogate ECU and extract data
	TIMESTAMP = time.time() # record initial timestamp for cycletime use		
	if ERRORCOUNT < ERRORMAX:
                time.sleep(TIMEUPDATE)

		DUMMY = serialw(REQTABLE11, TABLE11LEN) # Request table 11 data
		if DUMMY == "":
			ERRORCOUNT += 1
		else:
			ERRORCOUNT = 0
			ECU = DUMMY.split()
		
			RPM = int(ECU[5],16) * 256 + int(ECU[6],16)
			TPS = int(ECU[8],16)
			ECT = int(ECU[10],16) - 40
			IAT = int(ECU[12],16) - 40
			MAP = int(ECU[14],16)
			VAR15 = int(ECU[15],16)
			VAR16 = int(ECU[16],16)
			VOLTAGE = float(int(ECU[17],16))/10
			SPEED = int(ECU[18],16)
                        INJ = int(ECU[19],16) * 256 + int(ECU[20],16)
			VAR20 = int(ECU[20],16)
			VAR21 = int(ECU[21],16)
			VAR22 = int(ECU[22],16)
			VAR23 = int(ECU[23],16)
			VAR24 = int(ECU[24],16)

# ----------------------- Calculate gear
                        if SPEED > 5:
	                        RATIO = float(RPM) / float(SPEED)
                                for i in range(6,0,-1):
                                        if RATIO < float(GEARTABLE[i-1]):
                                                GEAR = i
                                                break
                        else:
                                GEAR = 0

# ---------------------- Calculate KLI
			if RPM * INJ >0:
				KLI = (1000 * KFUEL * SPEED / (INJ * RPM)) * ALPHA + KLI * (1 - ALPHA)
			if SPEED < 5:
				KLI = 0 # don't calculate km/l at low speed
			else:
				if KLI > KLA:
					KLA += BETA
				else:
					KLA -= BETA

				#KLA = KLI * BETA + KLA * ( 1 - BETA)

#----------------------- Calculate fuel delta
			try:
        			FILE = open(FUELFILE,'r')
        			FUEL = float(FILE.read())
        			FILE.close()
			except:
        			errorlog("Failed to read FUEL file")
        		TIMENOW = time.time()
			SCANTIME = TIMENOW - TIMESTAMP
			if FUEL <= FUELMIN:
				FUELDELTA += INJ * RPM * SCANTIME/(1000 * KFUEL * 3600) # fuel used since last update
			else:
				FUELDELTA = 0.0
                        #FILE = open(SCANTIMEFILE, 'w')
                        #FILE.write(str(SCANTIME))
                        #FILE.close

 			TIMESTAMP = TIMENOW  # update the timestamp

# ---------------------- Print output if req'd
		if PRINT:
			os.system('clear')
			print "RPM:", int(RPM)
			print "TPS:", TPS
			print "ECT:", ECT
			print "IAT:", IAT
			print "MAP:", MAP
			print "VLT:", '%.1f' % VOLTAGE
			print "SPD:", SPEED
			print "INJ:", INJ
			print "KLI:", KLI
			print "Gear:", GEAR
			try:
				print "Msg:", len(ECU)
			except:
				errorlog("ECU not defined yet")

# --------------------- Update disk files
                try:
                        FILE = open(INJECTORFILE, 'w')
                        FILE.write(str(INJ))
                        FILE.close
                except:
                        errorlog("Failed to write INJECTOR file")


                try:
                        FILE = open(FUELDELTAFILE, 'w')
                        FILE.write(str(FUELDELTA))
                        FILE.close
                except:
                        errorlog("Failed to write FUELDELTA file")

		
		try:
			FILE = open(VOLTFILE, 'w')
			FILE.write('%.1f' % VOLTAGE)
			FILE.close
		except:
			errorlog("Failed to write VOLTAGE file")

		try:
                	FILE = open(IATFILE, 'w')
                	FILE.write(str(IAT))
                	FILE.close
		except:
			errorlog("Failed to write IAT file")

		try:
                	FILE = open(ECTFILE, 'w')
                	FILE.write(str(ECT))
                	FILE.close
		except:
			errorlog("Failed to write ECT file")
		
		try:
                	FILE = open(KLIFILE, 'w')
                	FILE.write('%.1f' % KLI)
                	FILE.close
		except:
			errorlog("Failed to write KLI file")

		try:
			FILE = open(KLAFILE,'w')
			FILE.write('%.1f' % KLA)
			FILE.close()
		except:
			errorlog("Failed to write KLA file")
		try:
                	FILE = open(GEARFILE, 'w')
			if GEAR > 0:
                       		FILE.write(str(GEAR))
			else:
				FILE.write("-")
                	FILE.close
		except:
			errorlog("Failed to write GEAR file")
		
		try:
                	FILE = open(SPEEDFILE,'w')
                	FILE.write(str(SPEED))
                	FILE.close()
		except:
			errorlog("Failed to write SPEED file")

		try:
			FILE = open(RPMFILE,'w')
			FILE.write(str(RPM))
			FILE.close()
		except:
			errorlog("Failed to write RPM file")


	else:
		sys.exit(0)

